@@ -1,47 +1,36 @@
-{-# LANGUAGE CPP #-}
-{-# LANGUAGE ConstraintKinds #-}
 {-# LANGUAGE DataKinds #-}
-{-# LANGUAGE DeriveFunctor #-}
-{-# LANGUAGE ExistentialQuantification #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE GADTs #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE KindSignatures #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE PolyKinds #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TypeApplications #-}
 {-# LANGUAGE TypeFamilies #-}
-{-# LANGUAGE TypeOperators #-}
 {-# LANGUAGE UndecidableInstances #-}
--- fixme
-{-# OPTIONS_GHC -fno-warn-deprecations #-}
-{-# LANGUAGE DatatypeContexts #-}
+{-# OPTIONS_GHC -Wno-redundant-constraints #-}
+{-# OPTIONS_GHC -fno-warn-missing-signatures #-}
 
 module NumHask.Array where
 
-import Data.Distributive
-import Data.Functor.Rep
-import Data.Kind
+import Data.Distributive (Distributive(..))
+import Data.Functor.Rep (Representable(..), liftR2, pureRep, fmapRep)
 import Data.List ((!!))
-import Data.Promotion.Prelude
-import Data.Singletons as S
-import Data.Singletons.TypeLits as S
-import GHC.Exts
-import GHC.Show
+import GHC.Exts (IsList(..))
+import GHC.Show (Show(..))
+import NumHask.Error (impossible)
 import NumHask.Array.Constraints
+  (Fold, HeadModule, TailModule, IsValidConcat, Concatenate, Transpose, Squeeze)
 import NumHask.Prelude as P
-import NumHask.Shape
+import NumHask.Shape (HasShape(..))
 import Numeric.Dimensions as D
-import Numeric.Dimensions.XDim
 import qualified Data.Singletons.Prelude as S
 import qualified Data.Vector as V
-import qualified Protolude as Proto
-import qualified Test.QuickCheck as QC
 
--- $setup
+-- $setup 
 -- >>> :set -XDataKinds
 -- >>> :set -XOverloadedLists
 -- >>> :set -XTypeFamilies
@@ -60,7 +49,7 @@ import qualified Test.QuickCheck as QC
 data family Array (c :: Type -> Type) (ds :: [k]) (a :: Type)
 
 -- | instance where dimensions are known at compile time
-newtype instance (Dimensions ds) =>
+newtype instance
   Array c (ds :: [Nat]) t =
     Array { _getContainer :: c t}
     deriving (Functor, Foldable)
@@ -68,19 +57,25 @@ newtype instance (Dimensions ds) =>
 instance NFData (Array c ds t) where
   rnf a = seq a ()
 
+{-
 -- | instance of array where some of the dimensions are known at compile time
 -- it wraps an Array with some weird magic
 data instance Array c (xds :: [XNat]) t = forall (ds :: [Nat]).
-  ( FixedDim xds ds ~ ds
-  , FixedXDim xds ds ~ xds
+  ( FixedDims xds ds
   , Dimensions ds) =>
   SomeArray (Array c ds t)
 
+-}
+
+instance (Dimensions r) => HasShape (Array c (r :: [Nat])) where
+  type Shape (Array c r) = [Int]
+  shape _ = fmap fromIntegral (listDims $ dims @Nat @r)
+
 -- | an array with dimensions represented at the value level
 newtype AnyArray c a = AnyArray ([Int], c a)
 
 -- | convert an array with type-level shape to value-level shape
-anyArray :: (Dimensions ds) => Array c ds a -> AnyArray c a
+anyArray :: (Dimensions ds) => Array c (ds :: [Nat]) a -> AnyArray c a
 anyArray arr@(Array c) = AnyArray (shape arr, c)
 
 -- | a sweet class of container with attributes necessary to supply the set of operations here
@@ -123,25 +118,25 @@ instance Container [] where
   cfoldr = foldr
   cconcat = mconcat
 
-instance (Eq (c t), Dimensions ds) => Eq (Array c ds t) where
+instance (Eq (c t), Dimensions ds) => Eq (Array c (ds :: [Nat]) t) where
     (Array a) == (Array b) = a == b
 
-xdimList :: XDim ds -> [Int]
-xdimList (XDim d) = dimList d
-
-dimList :: Dim ds -> [Int]
-dimList D = []
+{-
+dimList :: Dims ds -> [Int]
+dimList U = []
 dimList (d :* ds) = dimList d ++ dimList ds
-dimList (Dn :: Dim m) = [dimVal' @m]
-dimList (Dx (Dn :: Dim m)) = [dimVal' @m]
+dimList (Dn _ :: Dim m) = [dimVal' @m]
+-- dimList (Dx (Dn _ :: Dim m)) = [dimVal' @m]
 
-instance (Dimensions r) => HasShape (Array c r) where
-  type Shape (Array c r) = [Int]
-  shape _ = dimList $ dim @r
+-}
 
+
+{-
 instance HasShape (Array c (xds :: [XNat])) where
   type Shape (Array c xds) = [Int]
   shape (SomeArray a) = shape a
+-}
+
 
 -- * shape helpers where dimensions ~ [Int]
 
@@ -167,33 +162,33 @@ unind ns x =
     ns
 
 instance forall r c. (Dimensions r, Container c) =>
-  Distributive (Array c r) where
-  distribute f = Array $ generate n $ \i -> fmap (\(Array v) -> idx v i) f
+  Data.Distributive.Distributive (Array c (r :: [Nat])) where
+  distribute f = Array $ generate (fromIntegral n) $ \i -> fmap (\(Array v) -> idx v i) f
     where
-      n = dimVal $ dim @r
+      n = totalDim $ dims @Nat @r
 
 instance forall r c. (Dimensions r, Container c) =>
-  Representable (Array c r) where
+  Representable (Array c (r :: [Nat])) where
   type Rep (Array c r) = [Int]
-  tabulate f = Array $ generate (product ns) (f . unind ns)
+  tabulate f = Array $ generate (fromIntegral $ product ns) (f . unind (fmap fromIntegral ns))
     where
-      ns = dimList $ dim @r
-  index (Array xs) rs = xs `idx` ind ns rs
+      ns = listDims $ dims @Nat @r
+  index (Array xs) rs = xs `idx` ind (fmap fromIntegral ns) rs
     where
-      ns = dimList $ dim @r
+      ns = listDims $ dims @Nat @r
 
 -- | from flat list
 instance
     ( Item (Array c r a) ~ Item (c a)
     , Dimensions r
-    , AdditiveUnital a
+    , Additive a
     , IsList (c a)
     ) =>
-    IsList (Array c r a) where
+    IsList (Array c (r :: [Nat]) a) where
   type Item (Array c r a) = a
   fromList l = Array $ fromList $ take n $ l ++ repeat zero
     where
-      n = dimVal (dim @r)
+      n = fromIntegral $ totalDim (dims @_ @r)
   toList (Array v) = GHC.Exts.toList v
 
 instance (Show a, Show (Item (c a)), Container c, IsList (c a)) => Show (AnyArray c a) where
@@ -221,22 +216,23 @@ flatten1 (AnyArray (rep, v)) =
         x:r -> (x, product r)
     ss = take n [0 ..]
 
-instance (Show a, Show (Item (c a)), IsList (c a), Container c, Dimensions ds) => Show (Array c ds a) where
+instance (Show a, Show (Item (c a)), IsList (c a), Container c, Dimensions ds)
+  => Show (Array c (ds :: [Nat]) a) where
   show = GHC.Show.show . anyArray
 
 type Vector c n = Array c '[ n]
 
 type Matrix c m n = Array c '[ m, n]
 
+{-
 instance
   ( IsList (c a)
   , Item (c a) ~ a
+  , Container c
   , KnownNat n
-  , AdditiveUnital (Vector c n a)
+  , Unital (Sum (Vector c n a))
   , QC.Arbitrary a
-  , AdditiveUnital a
-  , Num a
-  ) =>
+  , Additive a) =>
   QC.Arbitrary (Vector c n a) where
   arbitrary = QC.frequency [(1, pure zero), (9, fromList <$> QC.vector n)]
     where
@@ -245,19 +241,18 @@ instance
 instance
   ( IsList (c a)
   , Item (c a) ~ a
-  , AdditiveUnital (Matrix c m n a)
+  , Additive (Matrix c m n a)
+  , Container c
   , KnownNat m
   , KnownNat n
   , QC.Arbitrary a
-  , AdditiveUnital a
-  , Num a
-  ) =>
+  , Additive a) =>
   QC.Arbitrary (Matrix c m n a) where
   arbitrary = QC.frequency [(1, pure zero), (9, fromList <$> QC.vector (m * n))]
     where
       n = fromInteger $ P.natVal (Proxy :: Proxy n)
       m = fromInteger $ P.natVal (Proxy :: Proxy m)
-
+-}
 
 -- ** Operations
 -- | outer product
@@ -270,7 +265,7 @@ instance
 --  [3, 6, 9]]
 (><) :: forall c (r :: [Nat]) (s :: [Nat]) a.
   ( Container c
-  , CRing a
+  , CommutativeRing a
   , Dimensions r
   , Dimensions s
   , Dimensions ((D.++) r s))
@@ -298,26 +293,24 @@ instance
 --  [43, 50]]
 --
 mmult :: forall c m n k a.
-  ( Hilbert (Vector c k) a
+  ( Hilbert (Vector c k a)
   , Dimensions '[ m, k]
   , Dimensions '[ k, n]
   , Dimensions '[ m, n]
   , Container c
-  , KnownNat m
-  , KnownNat n
-  , KnownNat k
   )
-  => Matrix c m k a
+  => Matrix c (m :: Nat) (k :: Nat) a
   -> Matrix c k n a
   -> Matrix c m n a
-mmult x y = tabulate (\[i, j] -> unsafeRow i x <.> unsafeCol j y)
+mmult x y = tabulate go
+  where
+    go [i, j] = unsafeRow i x <.> unsafeCol j y
+    go _  = impossible "mmult only typechecks on arrays"
 
 -- | extract the row of a matrix
 row :: forall c i a m n.
   ( Dimensions '[ m, n]
   , Container c
-  , KnownNat m
-  , KnownNat n
   , KnownNat i
   , ((S.<) i m) ~ 'True
   )
@@ -326,24 +319,40 @@ row :: forall c i a m n.
   -> Vector c n a
 row i_ = unsafeRow i
   where
-    i = (Proto.fromIntegral . S.fromSing . S.singByProxy) i_
+    i = (fromIntegral . S.fromSing . S.singByProxy) i_
+
+rank2Shape
+  :: Dimensions '[ m, n]
+  => Matrix c (m :: Nat) (n :: Nat) a
+  -> (Int, Int)
+rank2Shape t =
+  case shape t of
+    [m, n] -> (m, n)
+    _      -> impossible "only typechecks for matricies"
 
 unsafeRow :: forall c a m n.
   ( Container c
   , Dimensions '[ m, n])
   => Int
-  -> Matrix c m n a
+  -> Matrix c (m :: Nat) (n :: Nat) a
   -> Vector c n a
 unsafeRow i t@(Array a) = Array $ cslice (i * n) n a
   where
-    [_, n] = shape t
+    (_, n) = rank2Shape t
+
+unsafeCol ::
+     forall c a m n. (Container c, Dimensions '[ m, n])
+  => Int
+  -> Matrix c (m :: Nat) (n :: Nat) a
+  -> Vector c m a
+unsafeCol j t@(Array a) = Array $ generate m (\x -> a `idx` (j + x * n))
+  where
+    (m, n) = rank2Shape t
 
 -- | extract the column of a matrix
 col :: forall c j a m n.
   ( Dimensions '[ m, n]
   , Container c
-  , KnownNat m
-  , KnownNat n
   , KnownNat j
   , ((S.<) j n) ~ 'True
   )
@@ -352,22 +361,14 @@ col :: forall c j a m n.
   -> Vector c m a
 col j_ = unsafeCol j
   where
-    j = (Proto.fromIntegral . S.fromSing . S.singByProxy) j_
+    j = (fromIntegral . S.fromSing . S.singByProxy) j_
 
-unsafeCol ::
-     forall c a m n. (Container c, Dimensions '[ m, n])
-  => Int
-  -> Matrix c m n a
-  -> Vector c m a
-unsafeCol j t@(Array a) = Array $ generate m (\x -> a `idx` (j + x * n))
-  where
-    [m, n] = shape t
 
 -- |
 --
 -- >>> unsafeIndex a [0,2,1]
 -- 10
-unsafeIndex :: (Container c, Dimensions r) => Array c r a -> [Int] -> a
+unsafeIndex :: (Container c, Dimensions r) => Array c (r :: [Nat]) a -> [Int] -> a
 unsafeIndex t@(Array a) i = a `idx` ind (shape t) i
 
 -- |
@@ -378,24 +379,10 @@ unsafeIndex t@(Array a) i = a `idx` ind (shape t) i
 unsafeSlice ::
      (Container c, IsList (c a), Item (c a) ~ a, Dimensions r, Dimensions r0)
   => [[Int]]
-  -> Array c r a
-  -> Array c r0 a
+  -> Array c (r :: [Nat]) a
+  -> Array c (r0 :: [Nat]) a
 unsafeSlice s t = Array (fromList [unsafeIndex t i | i <- sequence s])
 
--- | Slice xs = Map Length xs
-type family Slice (xss :: [[Nat]]) :: [Nat] where
-  Slice xss = Data.Promotion.Prelude.Map LengthSym0 xss
-
--- | AllLT xs n = All (n >) xs
-data AllLTSym0 (a :: S.TyFun [Nat] (S.TyFun Nat Bool -> Type))
-
-data AllLTSym1 (l :: [Nat]) (a :: S.TyFun Nat Bool)
-
-type instance S.Apply AllLTSym0 l = AllLTSym1 l
-
-type instance S.Apply (AllLTSym1 l) n =
-     Data.Promotion.Prelude.All ((S.>@#@$$) n) l
-
 -- |
 --
 -- todo: an ambiguous type variable has snuck in here somewhere
@@ -425,7 +412,7 @@ slice ::
 -}
 slice s_ = unsafeSlice s
   where
-    s = ((fmap . fmap) fromInteger . fromSing . singByProxy) s_
+    s = ((fmap . fmap) fromInteger . S.fromSing . S.singByProxy) s_
 
 -- |
 --
@@ -443,7 +430,7 @@ foldAlong ::
      , KnownNat s
      , Dimensions uvw
      , uw ~ (Fold s uvw)
-     , w ~ (Data.Promotion.Prelude.Drop 1 vw)
+     , w ~ (S.Drop 1 vw)
      , vw ~ (TailModule s uvw)
      )
   => Proxy s
@@ -460,7 +447,7 @@ foldAlong s_ f a@(Array v) =
        []
        md)
   where
-    s = (Proto.fromIntegral . fromSing . singByProxy) s_
+    s = (fromIntegral . S.fromSing . S.singByProxy) s_
     md = chunkItUp [] (product $ drop s $ shape a) v
 
 -- |
@@ -492,7 +479,7 @@ mapAlong s_ f a@(Array v) =
        []
        md)
   where
-    s = (Proto.fromIntegral . fromSing . singByProxy) s_
+    s = (fromIntegral . S.fromSing . S.singByProxy) s_
     md = chunkItUp [] (product $ drop s $ shape a) v
 
 -- |
@@ -508,7 +495,7 @@ mapAlong s_ f a@(Array v) =
 concatenate ::
      forall c s r t a.
      ( Container c
-     , SingI s
+     , S.SingI s
      , Dimensions r
      , Dimensions t
      , (IsValidConcat s t r) ~ 'True
@@ -520,7 +507,7 @@ concatenate ::
 concatenate s_ r@(Array vr) t@(Array vt) =
   Array . cconcat $ (concat . reverse . P.transpose) [rm, tm]
   where
-    s = (Proto.fromIntegral . fromSing . singByProxy) s_
+    s = (fromIntegral . S.fromSing . S.singByProxy) s_
     rm = chunkItUp [] (product $ drop s $ shape t) vt
     tm = chunkItUp [] (product $ drop s $ shape r) vr
 
@@ -542,8 +529,8 @@ concatenate s_ r@(Array vr) t@(Array vt) =
 --
 transpose ::
      forall c s t a. (t ~ Transpose s, Container c, Dimensions s, Dimensions t)
-  => Array c s a
-  -> Array c t a
+  => Array c (s :: [Nat]) a
+  -> Array c (t :: [Nat]) a
 transpose (Array x) = Array x
 
 -- |
@@ -588,146 +575,133 @@ squeeze ::
   -> Array c t a
 squeeze (Array x) = Array x
 
-instance (Dimensions r, Container c, AdditiveMagma a) =>
-         AdditiveMagma (Array c r a) where
-  plus = liftR2 plus
-
-instance (Dimensions r, Container c, AdditiveUnital a) =>
-         AdditiveUnital (Array c r a) where
+instance (Dimensions r, Container c, Additive a) =>
+  Additive (Array c (r :: [Nat]) a) where
+  a + b = liftR2 (+) a b
   zero = pureRep zero
 
-instance (Dimensions r, Container c, AdditiveAssociative a) =>
-         AdditiveAssociative (Array c r a)
-
-instance (Dimensions r, Container c, AdditiveCommutative a) =>
-         AdditiveCommutative (Array c r a)
-
-instance (Dimensions r, Container c, AdditiveInvertible a) =>
-         AdditiveInvertible (Array c r a) where
+instance (Dimensions r, Container c, Subtractive a) =>
+  Subtractive (Array c (r :: [Nat]) a) where
   negate = fmapRep negate
 
-instance (Dimensions r, Container c, Additive a) => Additive (Array c r a)
+instance (Dimensions r, Container c, Multiplicative a) =>
+  Multiplicative (Array c (r :: [Nat]) a) where
+  a * b = liftR2 (*) a b
 
-instance (Dimensions r, Container c, AdditiveGroup a) =>
-         AdditiveGroup (Array c r a)
-
-instance (Dimensions r, Container c, MultiplicativeMagma a) =>
-         MultiplicativeMagma (Array c r a) where
-  times = liftR2 times
-
-instance (Dimensions r, Container c, MultiplicativeUnital a) =>
-         MultiplicativeUnital (Array c r a) where
   one = pureRep one
 
-instance (Dimensions r, Container c, MultiplicativeAssociative a) =>
-         MultiplicativeAssociative (Array c r a)
-
-instance (Dimensions r, Container c, MultiplicativeCommutative a) =>
-         MultiplicativeCommutative (Array c r a)
-
-instance (Dimensions r, Container c, MultiplicativeInvertible a) =>
-         MultiplicativeInvertible (Array c r a) where
+instance (Dimensions r, Container c, Divisive a) =>
+  Divisive (Array c (r :: [Nat]) a) where
   recip = fmapRep recip
 
-instance (Dimensions r, Container c, Multiplicative a) =>
-         Multiplicative (Array c r a)
+instance (Dimensions r, Container c, Multiplicative a, Additive a) =>
+  P.Distributive (Array c (r :: [Nat]) a)
 
-instance (Dimensions r, Container c, MultiplicativeGroup a) =>
-         MultiplicativeGroup (Array c r a)
+instance (Dimensions r, Container c, IntegralDomain a) => IntegralDomain (Array c (r :: [Nat]) a)
 
-instance (Dimensions r, Container c, MultiplicativeMagma a, Additive a) =>
-         Distribution (Array c r a)
+instance (Dimensions r, Container c, Field a) => Field (Array c (r :: [Nat]) a)
 
-instance (Dimensions r, Container c, Semiring a) => Semiring (Array c r a)
-
-instance (Dimensions r, Container c, Ring a) => Ring (Array c r a)
-
-instance (Dimensions r, Container c, CRing a) => CRing (Array c r a)
-
-instance (Dimensions r, Container c, Field a) => Field (Array c r a)
-
-instance (Dimensions r, Container c, ExpField a) => ExpField (Array c r a) where
+instance (Dimensions r, Container c, ExpField a) => ExpField (Array c (r :: [Nat]) a) where
   exp = fmapRep exp
   log = fmapRep log
 
-instance (Foldable (Array c r), Dimensions r, Container c, BoundedField a) =>
-         BoundedField (Array c r a) where
-  isNaN f = or (fmapRep isNaN f)
+instance (Foldable (Array c r), Dimensions r, Container c, UpperBoundedField a) =>
+         UpperBoundedField (Array c (r :: [Nat]) a) where
+  isNaN = foldl' (||) False . fmapRep isNaN
 
-instance (Dimensions r, Container c, Signed a) => Signed (Array c r a) where
+instance (Foldable (Array c r), Dimensions r, Container c, LowerBoundedField a) =>
+         LowerBoundedField (Array c (r :: [Nat]) a)
+
+instance (Dimensions r, Container c, Multiplicative a, Signed a)
+  => Signed (Array c (r :: [Nat]) a) where
   sign = fmapRep sign
   abs = fmapRep abs
 
-instance (Functor (Array c r), Foldable (Array c r), ExpField a) =>
-         Normed (Array c r a) a where
-  size r = sqrt $ foldr (+) zero $ (** (one + one)) <$> r
+instance (Functor (Array c r), Foldable (Array c r), Additive (Array c r a), Normed a a, ExpField a) =>
+         Normed (Array c (r :: [Nat]) a) a where
+  normL1 r = foldr (+) zero $ normL1 <$> r
+  normL2 r = sqrt $ foldr (+) zero $ (** (one + one)) <$> r
 
-instance (Foldable (Array c r), Dimensions r, Container c, Epsilon a) =>
-         Epsilon (Array c r a) where
+instance (Eq (c a), Foldable (Array c r), Dimensions r, Container c, Epsilon a) =>
+         Epsilon (Array c (r :: [Nat]) a) where
+  epsilon = tabulate (const epsilon)
   nearZero f = and (fmapRep nearZero f)
   aboutEqual a b = and (liftR2 aboutEqual a b)
 
-instance (Foldable (Array c r), Dimensions r, Container c, ExpField a) =>
-         Metric (Array c r a) a where
-  distance a b = size (a - b)
+instance (Foldable (Array c r), Dimensions r, Container c, ExpField a, Subtractive a, Normed a a) =>
+         Metric (Array c (r :: [Nat]) a) a where
+  distanceL1 a b = normL1 (a - b)
+  distanceL2 a b = normL2 (a - b)
 
-instance (Dimensions r, Container c, Integral a) => Integral (Array c r a) where
+instance (Dimensions r, Container c, Integral a) => Integral (Array c (r :: [Nat]) a) where
   divMod a b = (d, m)
     where
       x = liftR2 divMod a b
       d = fmap fst x
       m = fmap snd x
+  quotRem a b = (q, r)
+    where
+      x = liftR2 quotRem a b
+      q = fmap fst x
+      r = fmap snd x
 
-instance (Foldable (Array c r), CRing a, Semiring a, Dimensions r, Container c) =>
-         Hilbert (Array c r) a where
-  a <.> b = sum $ liftR2 (*) a b
-
-instance (Dimensions r, Container c, Additive a) =>
-         AdditiveBasis (Array c r) a where
-  (.+.) = liftR2 (+)
-
-instance (Dimensions r, Container c, AdditiveGroup a) =>
-         AdditiveGroupBasis (Array c r) a where
-  (.-.) = liftR2 (-)
+type instance Actor (Array c r a) = a
 
 instance (Dimensions r, Container c, Multiplicative a) =>
-         MultiplicativeBasis (Array c r) a where
+  HadamardMultiplication (Array c (r :: [Nat])) a where
   (.*.) = liftR2 (*)
 
-instance (Dimensions r, Container c, MultiplicativeGroup a) =>
-         MultiplicativeGroupBasis (Array c r) a where
+instance (Dimensions r, Container c, Divisive a) =>
+  HadamardDivision (Array c (r :: [Nat])) a where
   (./.) = liftR2 (/)
 
 instance (Dimensions r, Container c, Additive a) =>
-         AdditiveModule (Array c r) a where
+  AdditiveAction (Array c (r::[Nat]) a) where
   (.+) r s = fmap (s +) r
   (+.) s = fmap (s +)
 
-instance (Dimensions r, Container c, AdditiveGroup a) =>
-         AdditiveGroupModule (Array c r) a where
+instance (Dimensions r, Container c, Subtractive a) =>
+  SubtractiveAction (Array c (r::[Nat]) a) where
   (.-) r s = fmap (\x -> x - s) r
   (-.) s = fmap (\x -> x - s)
 
 instance (Dimensions r, Container c, Multiplicative a) =>
-         MultiplicativeModule (Array c r) a where
-  (.*) r s = fmap (s *) r
+  MultiplicativeAction (Array c (r :: [Nat]) a) where
+  (.*) r s = fmap (* s) r
   (*.) s = fmap (s *)
 
-instance (Dimensions r, Container c, MultiplicativeGroup a) =>
-         MultiplicativeGroupModule (Array c r) a where
+instance (Dimensions r, Container c, Divisive a) =>
+  DivisiveAction (Array c (r::[Nat]) a) where
   (./) r s = fmap (/ s) r
   (/.) s = fmap (/ s)
 
-instance (Dimensions r, Container c) => Singleton (Array c r) where
-  singleton = pureRep
+instance forall a c r. (Actor (Array c r a) ~ a, Foldable (Array c r), P.Distributive a, CommutativeRing a, Semiring a, Dimensions r, Container c) =>
+  Hilbert (Array c (r :: [Nat]) a) where
+  a <.> b = sum $ liftR2 (*) a b
 
-instance ( Foldable (Array c r)
-         , Dimensions r
-         , Container c
-         , CRing a
-         , Multiplicative a
-         ) =>
-         TensorProduct (Array c r a) where
+instance
+  ( Foldable (Array c r)
+  , Dimensions r
+  , Container c
+  , CommutativeRing a
+  , Multiplicative a
+  ) =>
+  TensorProduct (Array c (r :: [Nat]) a) where
   (><) m n = tabulate (\i -> index m i *. n)
   timesleft v m = tabulate (\i -> v <.> index m i)
   timesright m v = tabulate (\i -> v <.> index m i)
+
+instance (Eq (c a), Container c, Dimensions r, JoinSemiLattice a) => JoinSemiLattice (Array c (r :: [Nat]) a) where
+  (\/) = liftR2 (\/)
+
+instance (Eq (c a), Container c, Dimensions r, MeetSemiLattice a) => MeetSemiLattice (Array c (r :: [Nat]) a) where
+  (/\) = liftR2 (/\)
+
+instance (Eq (c a), Container c, Dimensions r, BoundedJoinSemiLattice a) => BoundedJoinSemiLattice (Array c (r :: [Nat]) a) where
+  bottom = pureRep bottom
+
+instance (Eq (c a), Container c, Dimensions r, BoundedMeetSemiLattice a) => BoundedMeetSemiLattice (Array c (r :: [Nat]) a) where
+  top = pureRep top
+
+singleton :: (Dimensions r, Container c) => a -> Array c (r :: [Nat]) a
+singleton a = tabulate (const a)
